theory Demo10
  imports Main
  abbrevs
    \<comment> \<open>For more convenient input of bold symbols\<close>
    "b0" = "\<^bold>0" and
    "b+" = "\<^bold>+" and
    "b*" = "\<^bold>*" and
    "b-" = "\<^bold>-" and
    "b<" = "\<^bold><"
begin

section \<open>Notation\<close>

text \<open>
See also Chapter 8.2 of @{doc "isar-ref"}.
\<close>

subsection \<open>Arithmetic Expressions\<close>

datatype aexp = Zero | Plus aexp aexp | Times aexp aexp | Minus aexp

notation
  Zero ("\<^bold>0" [] 1000) and
  Plus (infixl "\<^bold>+" 0) and
  Times (infixr "\<^bold>*" 2) and
  Minus ("\<^bold>-_" [3] 3)

lemma "(\<^bold>0 \<^bold>+ \<^bold>0 \<^bold>* \<^bold>0) = Plus Zero (Times Zero Zero)" ..




subsection \<open>Haskell-Like List Syntax\<close>

text \<open>
In Haskell

  \<^item> \<open>[]\<close> is used for the empty list,
  \<^item> \<open>x:xs\<close> for the non-empty list with head \<open>x\<close> and tail \<open>xs\<close>, and
  \<^item> \<open>[x\<^sub>1,x\<^sub>2,\<dots>,x\<^sub>n]\<close> to enumerate lists with specific elements.

Before we can introduce Haskell-like syntax, we have to get rid of existing notation that
would clash.
\<close>

\<comment> \<open>Copied from @{theory HOL.Set}\<close>
no_notation (ASCII)
  Set.member  ("'(:')") and
  Set.member  ("(_/ : _)" [51, 51] 50) and
  Set.not_member  ("'(~:')") and
  Set.not_member  ("(_/ ~: _)" [51, 51] 50)

\<comment> \<open>Copied from @{theory HOL.List}\<close>
no_notation Nil ("[]")

text \<open>
Now we can introduce a new data type for lists with Haskell-like constructor syntax.
\<close>
datatype 'a LIST = NIL ("[]") | CONS 'a "'a LIST" (infixr ":" 65)

term "x:y:z:[]"

text \<open>
In order to support finite enumerations, we need to employ syntax translations.
\<close>

\<comment> \<open>Copied from @{theory HOL.List}\<close>
no_syntax
  \<comment> \<open>list Enumeration\<close>
  "_list" :: "args => 'a list"    ("[(_)]")

nonterminal elts

syntax
  "" :: "logic \<Rightarrow> elts" ("_")
  "_elts" :: "logic \<Rightarrow> elts \<Rightarrow> elts" ("_,/ _")

syntax
  "_LIST" :: "elts \<Rightarrow> 'a LIST" ("[(_)]")

translations
  "[x, xs]" \<rightleftharpoons> "x : [xs]"
  "[x]" \<rightleftharpoons> "x : []"

term "[1,2,3,4]"

\<comment> \<open>Copied from @{theory HOL.Map}\<close>
no_notation
  map_add  (infixl "++" 100)

fun append :: "'a LIST \<Rightarrow> 'a LIST \<Rightarrow> 'a LIST" (infixr "++" 65)
  where
    "[] ++ ys = ys"
  | "(x:xs) ++ ys = x : (xs ++ ys)"




section \<open>Type Classes\<close>

text \<open>
See also @{doc classes}.
\<close>

subsection \<open>Transforming Arbitrary Values to Strings\<close>

type_synonym STRING = "char LIST"

text \<open>Transforming ``our'' lists to standard lists.\<close>
fun to_list :: "'a LIST \<Rightarrow> 'a list"
  where
    "to_list [] = Nil"
  | "to_list (x:xs) = Cons x (to_list xs)"

class to_string =
  fixes to_string :: "'a \<Rightarrow> STRING"
begin

abbreviation "show \<equiv> to_list \<circ> to_string"

end

declare [[show_sorts]]
term "show"
declare [[show_sorts=false]]

text \<open>
Characters are an instance of the @{class to_string} class, that is,
they can be transformed into @{typ STRING}s
\<close>
instantiation char :: to_string
begin

definition to_string_char :: "char \<Rightarrow> STRING"
  where
    "to_string_char c = [c]"

instance ..

end

text \<open>
Pairs are an instance of the @{class to_string} class.
\<close>
instantiation prod :: (to_string, to_string) to_string
begin

fun to_string_prod :: "('a \<times> 'b) \<Rightarrow> STRING"
  where
    "to_string_prod (x, y) = CHR ''('' : to_string x ++ CHR '','' : to_string y ++ [CHR '')'']"

instance ..

end

value "show (CHR ''c'', CHR ''d'')"

text \<open>
Lists are an instance of @{class to_string}.
\<close>
instantiation list :: (to_string) to_string
begin

fun to_string_list :: "'a list \<Rightarrow> STRING"
  where
    "to_string_list Nil = []"
  | "to_string_list (x#xs) = to_string x ++ to_string_list xs"

instance ..

end

value "show ''foo bar?''"


subsection \<open>Total Orders\<close>

class partial_order =
  fixes LEQ :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<^bold>\<le>" 50)
  assumes refl: "x \<^bold>\<le> x"
    and antisym: "x \<^bold>\<le> y \<Longrightarrow> y \<^bold>\<le> x \<Longrightarrow> x = y"
    and trans: "x \<^bold>\<le> y \<Longrightarrow> y \<^bold>\<le> z \<Longrightarrow> x \<^bold>\<le> z"
begin

definition LT :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<^bold><" 50)
  where
    "x \<^bold>< y \<longleftrightarrow> x \<^bold>\<le> y \<and> x \<noteq> y"

end

class total_order = partial_order +
  assumes total: "x \<^bold>< y \<or> y \<^bold>< x \<or> x = y"
begin

end

instantiation nat :: partial_order
begin

fun LEQ_nat :: "nat \<Rightarrow> nat \<Rightarrow> bool"
  where
    "0 \<^bold>\<le> (y::nat) \<longleftrightarrow> True"
  | "Suc x \<^bold>\<le> Suc y \<longleftrightarrow> x \<^bold>\<le> y"
  | "x \<^bold>\<le> (y::nat) \<longleftrightarrow> False"

lemma LEQ_nat_def:
  fixes x y :: nat
  shows "x \<^bold>\<le> y \<longleftrightarrow> (\<exists>n. y = x + n)"
  by (induction x y rule: LEQ_nat.induct) auto

instance by (standard) (auto simp: LEQ_nat_def)

end

lemma LEQ_nat_total:
  fixes x y :: nat
  shows "x \<^bold>\<le> y \<or> y \<^bold>\<le> x"
proof (induction x arbitrary: y)
  case (Suc x)
  then show ?case
    by (cases y) auto
qed simp

instance nat :: total_order
proof
  fix x y :: nat
  show "x \<^bold>< y \<or> y \<^bold>< x \<or> x = y"
    using LEQ_nat_total [of x y]
    unfolding LT_def by blast
qed

end
